from __future__ import annotations
from typing import Any
import energizer.tensor as ts
import numpy as np
try:
    import mlx.core as mx
except ImportError:
    mx = None

def _reduce_grad_to_shape(grad: Any, shape: tuple) -> Any:
    if shape == ():
        if mx and isinstance(grad, mx.array):
            return mx.sum(grad)
        return np.sum(grad)

    is_mlx = mx and isinstance(grad, mx.array)

    g = grad
    while len(getattr(g, "shape", ())) > len(shape):
        if is_mlx:
            g = mx.sum(g, axis=0)
        else:
            g = np.sum(g, axis=0)

    for axis, (gdim, tdim) in enumerate(zip(g.shape, shape)):
        if tdim == 1 and gdim != 1:
            if is_mlx:
                g = mx.sum(g, axis=axis, keepdims=True)
            else:
                g = np.sum(g, axis=axis, keepdims=True)

    try:
        return g.reshape(shape)
    except Exception:
        return g

def mul_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
    
    def get_data(t):
        """Extract raw data from tensor or return as-is"""
        if isinstance(t, (int, float)):
            return t
        elif isinstance(t, ts.Tensor):
            return t.data
        else:
            return t
    
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        other_data = get_data(tensors[1])
        if mx and (isinstance(grad_data, mx.array) or isinstance(other_data, mx.array)):
            grad_mx = grad_data if isinstance(grad_data, mx.array) else mx.array(grad_data)
            other_mx = other_data if isinstance(other_data, mx.array) else mx.array(other_data) if isinstance(other_data, np.ndarray) else other_data
            tensors[0].grad = np.array(grad_mx * other_mx)
        else:
            tensors[0].grad = grad_data * other_data
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        other_data = get_data(tensors[0])
        if mx and (isinstance(grad_data, mx.array) or isinstance(other_data, mx.array)):
            grad_mx = grad_data if isinstance(grad_data, mx.array) else mx.array(grad_data)
            other_mx = other_data if isinstance(other_data, mx.array) else mx.array(other_data) if isinstance(other_data, np.ndarray) else other_data
            tensors[1].grad = np.array(grad_mx * other_mx)  # Store as numpy for .grad
        else:
            tensors[1].grad = grad_data * other_data
        tensors[1].backward(tensors[1].grad)
    else:
        return grad_data * get_data(tensors[1])
    return grad_data * get_data(tensors[0])

def add_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        grad0 = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
        tensors[0].grad = _reduce_grad_to_shape(grad0, tensors[0].data.shape)
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        grad1 = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
        tensors[1].grad = _reduce_grad_to_shape(grad1, tensors[1].data.shape)
        tensors[1].backward(tensors[1].grad)
    else:
        return grad_outputs[0]
    return grad_outputs[0]

def sub_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        grad0 = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
        tensors[0].grad = _reduce_grad_to_shape(grad0, tensors[0].data.shape)
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        grad1 = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
        tensors[1].grad = _reduce_grad_to_shape(-grad1, tensors[1].data.shape)
        tensors[1].backward(tensors[1].grad)
    else:
        return -grad_outputs[0]
    return grad_outputs[0]

def neg_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = -grad_outputs[0] if isinstance(grad_outputs[0], (int, float)) else -grad_outputs[0].data
        tensors[0].backward(tensors[0].grad)
    else:
        return -grad_outputs[0] if isinstance(grad_outputs[0], (int, float)) else -grad_outputs[0].data
    return grad_outputs[0]

def truediv_backward(tensors: Any, grad_outputs: Any) -> Any:
    is_reverse_div = isinstance(tensors[0], (int, float)) and isinstance(tensors[1], ts.Tensor)

    if is_reverse_div:
        if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
            tensors[1].grad = -grad_outputs[0] * tensors[0] / (tensors[1].data ** 2)
            tensors[1].backward(tensors[1].grad)
        return grad_outputs[0]

    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_outputs[0] / (tensors[1] if isinstance(tensors[1], (int, float)) else tensors[1].data)
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        tensors[1].grad = -grad_outputs[0] * tensors[0].data / (tensors[1].data ** 2)
        tensors[1].backward(tensors[1].grad)
    return grad_outputs[0]

def matmul_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
    
    def ensure_compatible(a, b):
        if mx and isinstance(a, mx.array) and not isinstance(b, mx.array):
            return a, mx.array(b)
        elif mx and isinstance(b, mx.array) and not isinstance(a, mx.array):
            return mx.array(a), b
        return a, b
    
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        grad_compat, weight_T = ensure_compatible(grad_data, tensors[1].data.T)
        tensors[0].grad = grad_compat @ weight_T
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        input_T, grad_compat = ensure_compatible(tensors[0].data.T, grad_data)
        tensors[1].grad = input_T @ grad_compat
        tensors[1].backward(tensors[1].grad)
    
    grad_compat, weight_T = ensure_compatible(grad_data, tensors[1].data.T)
    return grad_compat @ weight_T

def pow_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_outputs[0] * tensors[1] * (tensors[0].data ** (tensors[1] - 1))
        tensors[0].backward(tensors[0].grad)
    return grad_outputs[0] * tensors[1] * (tensors[0].data ** (tensors[1] - 1))

def max_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_outputs[0] * (tensors[0].data > tensors[1])
        tensors[0].backward(tensors[0].grad)
    return grad_outputs[0] * (tensors[0].data > tensors[1])

def sum_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
    
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        if not isinstance(grad_data, np.ndarray):
            grad_data = np.array(grad_data)
        tensors[0].grad = np.broadcast_to(grad_data, tensors[0].data.shape).copy()
        tensors[0].backward(tensors[0].grad)
    return grad_data

def mean_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = np.broadcast_to(grad_data / tensors[0].data.size, tensors[0].data.shape).copy()
        tensors[0].backward(tensors[0].grad)
    return grad_data / tensors[0].data.size

def sigmoid_backward(tensors: Any, grad_outputs: Any) -> Any:
    """
    Backward for sigmoid(x).

    `tensors` is expected to be [x, y] where y = sigmoid(x) (raw array).
    """
    x = tensors[0]
    y = tensors[1]
    grad = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]

    # dy/dx = y * (1 - y)
    if mx and (isinstance(grad, mx.array) or isinstance(y, mx.array)):
        grad_mx = grad if isinstance(grad, mx.array) else mx.array(grad)
        y_mx = y if isinstance(y, mx.array) else mx.array(y)
        grad_x = grad_mx * (y_mx * (1 - y_mx))
    else:
        grad_np = np.array(grad) if not isinstance(grad, np.ndarray) else grad
        y_np = np.array(y) if not isinstance(y, np.ndarray) else y
        grad_x = grad_np * (y_np * (1 - y_np))

    if isinstance(x, ts.Tensor) and x.requires_grad:
        x.grad = grad_x
        x.backward(x.grad)

    return grad_x

def as_strided_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = np.lib.stride_tricks.as_strided(grad_outputs[0], shape=tensors[0].data.shape, strides=tensors[1].data.strides, storage_offset=tensors[2].data)
        tensors[0].backward(tensors[0].grad)
    return grad_outputs[0]

def trace_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = np.trace(grad_outputs[0])
        tensors[0].backward(tensors[0].grad)
    return np.trace(grad_outputs[0])

def conv1d_backward(tensors: Any, grad_outputs: Any, stride: int = 1, padding: int = 0) -> Any:
    x, weight, bias = tensors
    grad_output = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
    
    if not isinstance(grad_output, np.ndarray):
        grad_output = np.array(grad_output)
    
    weight_data = weight.data if isinstance(weight, ts.Tensor) else weight
    
    if isinstance(x, ts.Tensor) and x.requires_grad:
        grad_x = conv1d_grad_input(grad_output, weight_data, stride, padding)
        if grad_x is not None:
            x.grad = grad_x
            x.backward(x.grad)
    
    if isinstance(weight, ts.Tensor) and weight.requires_grad:
        grad_weight = conv1d_grad_weight(grad_output, x, stride, padding)
        if grad_weight is not None:
            weight.grad = grad_weight
            weight.backward(weight.grad)
    
    if bias is not None and isinstance(bias, ts.Tensor) and bias.requires_grad:
        grad_bias = grad_output.sum(axis=(0, 2))
        bias.grad = grad_bias
        bias.backward(bias.grad)
    
    return grad_output

def conv1d_grad_input(grad_output: np.ndarray, weight: 'ts.Tensor', 
                      stride: int = 1, padding: int = 0) -> np.ndarray:
    batch_size, out_channels, out_length = grad_output.shape
    _, in_channels, kernel_size = weight.data.shape
    
    length = (out_length - 1) * stride + kernel_size - 2 * padding
    
    grad_input_padded = np.zeros((batch_size, in_channels, length + 2 * padding))
    
    for b in range(batch_size):
        for oc in range(out_channels):
            for ol in range(out_length):
                il_start = ol * stride
                
                for k in range(kernel_size):
                    grad_input_padded[b, :, il_start + k] += \
                        weight[oc, :, k] * grad_output[b, oc, ol]
    
    if padding > 0:
        return grad_input_padded[:, :, padding:-padding]
    return grad_input_padded


def conv1d_grad_weight(grad_output: np.ndarray, x: 'ts.Tensor',
                       stride: int = 1, padding: int = 0) -> np.ndarray:
    batch_size, out_channels, out_length = grad_output.shape
    _, in_channels, _ = x.data.shape
    
    kernel_size = x.data.shape[2] - (out_length - 1) * stride + 2 * padding
    
    grad_weight = np.zeros((out_channels, in_channels, kernel_size))
    
    if padding > 0:
        x_padded = np.pad(x.data, 
                         ((0, 0), (0, 0), (padding, padding)), 
                         mode='constant')
    else:
        x_padded = x.data
    
    for b in range(batch_size):
        for oc in range(out_channels):
            for ol in range(out_length):
                il_start = ol * stride
                il_end = il_start + kernel_size
                
                x_window = x_padded[b, :, il_start:il_end]
                
                grad_weight[oc] += x_window * grad_output[b, oc, ol]
    
    return grad_weight

def reshape_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
    
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        if mx and isinstance(grad_data, mx.array):
            tensors[0].grad = mx.reshape(grad_data, tensors[0].data.shape)
        else:
            if not isinstance(grad_data, np.ndarray):
                grad_data = np.array(grad_data)
            tensors[0].grad = grad_data.reshape(tensors[0].data.shape)
        tensors[0].backward(tensors[0].grad)
    
    if mx and isinstance(grad_data, mx.array):
        return mx.reshape(grad_data, tensors[0].data.shape)
    else:
        if not isinstance(grad_data, np.ndarray):
            grad_data = np.array(grad_data)
        return grad_data.reshape(tensors[0].data.shape)

def getitem_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = np.zeros_like(tensors[0].data)
        tensors[0].grad[tensors[1]] = grad_outputs[0]
        tensors[0].backward(tensors[0].grad)
    return np.zeros_like(tensors[0].data)

def setitem_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad[tensors[1]] = grad_outputs[0]
        tensors[0].backward(tensors[0].grad)
    return np.zeros_like(tensors[0].data)

def item_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_outputs[0]
        tensors[0].backward(tensors[0].grad)
    return grad_outputs[0]

def transpose_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data.T
    else:
        try:
            import mlx.core as mx
            if isinstance(grad_data, mx.array):
                tensors[0].grad = mx.transpose(grad_data)
            else:
                tensors[0].grad = np.transpose(grad_data)
        except ImportError:
            tensors[0].grad = np.transpose(grad_data)
    return tensors[0].grad.T if hasattr(tensors[0], 'grad') else tensors[0].grad

def tanh_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_outputs[0] * (1 - tensors[0].data ** 2)
        tensors[0].backward(tensors[0].grad)
    return grad_outputs[0] * (1 - tensors[0].data ** 2)

def softmax_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_outputs[0] * (tensors[0].data > 0)
        tensors[0].backward(tensors[0].grad)
    return grad_outputs[0] * (tensors[0].data > 0)

def transpose_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = grad_outputs[0].data if isinstance(grad_outputs[0], ts.Tensor) else grad_outputs[0]
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data.T
    else:
        try:
            import mlx.core as mx
            if isinstance(grad_data, mx.array):
                tensors[0].grad = mx.transpose(grad_data)
            else:
                tensors[0].grad = np.transpose(grad_data)
        except ImportError:
            tensors[0].grad = np.transpose(grad_data)
    return tensors[0].grad.T if hasattr(tensors[0], 'grad') else tensors[0].grad


from ctypes import Union
from energizer.tensor import Tensor
import numpy as np
import numpy.lib.stride_tricks as np_stride_tricks
import mlx.core as mx

from energizer.function import Function
import energizer.derivatives as dv

def max(tensor : Tensor, floor: Union[int, float] = 0) -> Tensor:
    if tensor.device == 'gpu':
        return Tensor(mx.maximum(floor, tensor.data), requires_grad=tensor.requires_grad, grad_fn=Function(dv.max_backward, [tensor, floor]), device=tensor.device)
    return Tensor(np.maximum(floor, tensor.data), requires_grad=tensor.requires_grad, grad_fn=Function(dv.max_backward, [tensor, floor]), device=tensor.device)

def as_strided(tensor: Tensor, shape: tuple, strides: tuple, storage_offset: int = 0) -> 'Tensor':
    array = tensor.data

    itemsize = array.itemsize
    byte_strides = tuple(s * itemsize for s in strides)
    
    if storage_offset > 0:
        flat_array = array.flatten()
        if storage_offset >= len(flat_array):
            raise ValueError(f"storage_offset {storage_offset} is out of bounds for array of size {len(flat_array)}")
        offset_array = flat_array[storage_offset:]
        array = offset_array

    if tensor.device == 'gpu':
        strided_data = mx.as_strided(array, shape=shape, strides=byte_strides, storage_offset=storage_offset)
    else:
        strided_data = np_stride_tricks.as_strided(array, shape=shape, strides=byte_strides)

    return Tensor(strided_data, requires_grad=tensor.requires_grad, grad_fn=Function(dv.as_strided_backward, [tensor, shape, strides, storage_offset]), device=tensor.device)

def trace(tensor: Tensor) -> Tensor:
    if tensor.device == 'gpu':
        result = mx.trace(tensor.data)
    else:
        result = np.trace(tensor.data)
    return Tensor(result, requires_grad=tensor.requires_grad, grad_fn=Function(dv.trace_backward, [tensor]), device=tensor.device)

def tanh(tensor: Tensor) -> Tensor:
    if tensor.device == 'gpu':
        return Tensor(mx.tanh(tensor.data), requires_grad=tensor.requires_grad, grad_fn=Function(dv.tanh_backward, [tensor]), device=tensor.device)
    return Tensor(np.tanh(tensor.data), requires_grad=tensor.requires_grad, grad_fn=Function(dv.tanh_backward, [tensor]), device=tensor.device)

def softmax(tensor: Tensor) -> Tensor:
    if tensor.device == 'gpu':
        return Tensor(mx.softmax(tensor.data), requires_grad=tensor.requires_grad, grad_fn=Function(dv.softmax_backward, [tensor]), device=tensor.device)
    return Tensor(np.softmax(tensor.data), requires_grad=tensor.requires_grad, grad_fn=Function(dv.softmax_backward, [tensor]), device=tensor.device)


from energizer.neural_network import Module
from energizer.tensor import Tensor
import numpy as np

class MSELoss(Module):
    def __init__(self, size_average: bool = None, reduce: bool = None, reduction: str = 'mean'):
        super().__init__()
        self.size_average = size_average
        self.reduce = reduce
        self.reduction = reduction

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        diff = (input - target) ** 2
        if self.reduction == 'sum':
            return diff.sum()
        return diff.mean()

    def __call__(self, input: Tensor, target: Tensor) -> Tensor:
        return self.forward(input, target)

class CrossEntropyLoss(Module):
    def __init__(self, size_average: bool = None, reduce: bool = None, reduction: str = 'mean'):
        super().__init__()
        self.size_average = size_average
        self.reduce = reduce
        self.reduction = reduction

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return -target * np.log(input)

    def __call__(self, input: Tensor, target: Tensor) -> Tensor:
        return self.forward(input, target)


from energizer.function import Function
from energizer.neural_network import Module, Parameter
from energizer.tensor import Tensor
import numpy as np
import mlx.core as mx
import energizer.functionnal as F
import energizer.derivatives as dv

class ConvNd(Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int, stride: int = 1, padding: int = 0, bias: bool = True):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        if self.device == 'gpu':
            std = mx.sqrt(2.0 / (in_channels * kernel_size))
            self.weight = Parameter(mx.random.normal(shape=(out_channels, in_channels, kernel_size)), requires_grad=True, device=self.device)
        else:
            std = np.sqrt(2.0 / (in_channels * kernel_size))
            self.weight = Parameter(np.random.randn(out_channels, in_channels, kernel_size) * std, requires_grad=True, device=self.device)
        if bias:
            if self.device == 'gpu':
                self.bias = Parameter(mx.zeros(out_channels), requires_grad=True, device=self.device)
            else:
                self.bias = Parameter(np.zeros(out_channels), requires_grad=True, device=self.device)
        else:
            self.bias = None
        self._parameters['weight'] = self.weight
        if self.bias is not None:
            self._parameters['bias'] = self.bias

    def forward(self) -> Tensor:
        pass


class Conv1d(ConvNd):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int, stride: int = 1, padding: int = 0, bias: bool = True):
        super().__init__(in_channels, out_channels, kernel_size, stride, padding, bias)

    def forward(self, input: Tensor) -> Tensor:
        self._input = input
        self._input_padded = self._pad1d(input, self.padding) if self.padding > 0 else input

        batch_size, in_channels, length = self._input_padded.data.shape
        out_length = ((length - self.kernel_size) // self.stride) + 1

        x_2d = self._input_padded.data.reshape(-1, length)

        windows = []
        for i in range(x_2d.shape[0]):
            row_windows = F.as_strided(Tensor(x_2d[i]), shape=(out_length, self.kernel_size), strides=(self.stride, 1)).data
            windows.append(row_windows)

        windows_stacked = np.stack(windows, axis=0)

        windows_reshaped = windows_stacked.reshape(
            batch_size, in_channels, out_length, self.kernel_size).transpose(0, 2, 1, 3).reshape(batch_size, out_length, -1
        )

        weight_reshaped = self.weight.data.reshape(self.out_channels, -1)

        result = np.einsum('bij,oj->bio', windows_reshaped, weight_reshaped)
        result = result.transpose(0, 2, 1)

        if self.bias is not None:
            result += self.bias.data.reshape(1, -1, 1)

        return Tensor(result, requires_grad=input.requires_grad, grad_fn=Function(dv.conv1d_backward, [self._input, self.weight, self.bias], stride=self.stride, padding=self.padding))

    def _pad1d(self, input: Tensor, padding: int) -> Tensor:
        batch_size, channels, length = input.data.shape
        padded_length = length + 2 * padding
        padded_data = np.zeros((batch_size, channels, padded_length))
        padded_data[:, :, padding:padding+length] = input.data
        return Tensor(padded_data, requires_grad=input.requires_grad)

    def _unfold1d(self, input: Tensor, kernel_size: int, stride: int, out_length: int) -> Tensor:
        batch_size, channels, length = input.data.shape
        unfolded_data = np.zeros((batch_size, out_length, channels, kernel_size))
        for b in range(batch_size):
            for c in range(channels):
                channel_data = input.data[b, c, :]

                window = F.as_strided(Tensor(channel_data), shape=(out_length, kernel_size), strides=(stride, 1))

                unfolded_data[b, :, c, :] = window.data

        return Tensor(unfolded_data, requires_grad=input.requires_grad)

    def to(self, device: str):
        self.weight = self.weight.to(device)
        if self.bias is not None:
            self.bias = self.bias.to(device)
        return self

class Conv2d(Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: tuple, stride: tuple = (1, 1), padding: tuple = (0, 0), bias: bool = True):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
        self.stride = stride if isinstance(stride, tuple) else (stride, stride)
        self.padding = padding if isinstance(padding, tuple) else (padding, padding)

        if self.device == 'gpu':
            std = mx.sqrt(2.0 / (in_channels * self.kernel_size[0] * self.kernel_size[1]))
            self.weight = Parameter(mx.random.normal(shape=(out_channels, in_channels, self.kernel_size[0], self.kernel_size[1])), requires_grad=True, device=self.device)
        else:
            std = np.sqrt(2.0 / (in_channels * self.kernel_size[0] * self.kernel_size[1]))
            self.weight = Parameter(np.random.randn(out_channels, in_channels, self.kernel_size[0], self.kernel_size[1]) * std, requires_grad=True, device=self.device)    
        if bias:
            if self.device == 'gpu':
                self.bias = Parameter(mx.zeros(out_channels), requires_grad=True, device=self.device)
            else:
                self.bias = Parameter(np.zeros(out_channels), requires_grad=True, device=self.device)
        else:
            self.bias = None

    def forward(self, input: Tensor) -> Tensor:
        if self.padding != (0, 0):
            input = self._pad2d(input, self.padding)

        batch_size, in_channels, height, width = input.data.shape
        kh, kw = self.kernel_size
        sh, sw = self.stride

        out_height = ((height - kh) // sh) + 1
        out_width = ((width - kw) // sw) + 1

        windows = self._im2col(input, kh, kw, sh, sw, out_height, out_width)

        if input.device == 'gpu':
            weight_reshaped = input.mlx().__class__(self.weight.data.reshape(self.out_channels, -1))

            result = mx.zeros((batch_size, out_height * out_width, self.out_channels))
            for b in range(batch_size):
                result[b] = windows[b] @ weight_reshaped.T
            result = result.transpose(0, 2, 1).reshape(batch_size, self.out_channels, out_height, out_width)
            if self.bias is not None:
                result += self.bias.data.reshape(1, -1, 1, 1)
        else:
            weight_reshaped = self.weight.data.reshape(self.out_channels, -1)   

            result = np.zeros((batch_size, out_height * out_width, self.out_channels))
            for b in range(batch_size):
                result[b] = windows[b] @ weight_reshaped.T

            result = result.transpose(0, 2, 1).reshape(batch_size, self.out_channels, out_height, out_width)

            if self.bias is not None:
                result += self.bias.data.reshape(1, -1, 1, 1)
            
        return Tensor(result, requires_grad=input.requires_grad, device=input.device)


    def _pad2d(self, x: Tensor, padding: tuple) -> Tensor:
        ph, pw = padding
        batch_size, channels, height, width = x.data.shape
        if ph == 0 and pw == 0:
            return x
        padded_data = np.zeros((batch_size, channels, height + 2*ph, width + 2*pw))
        padded_data[:, :, ph:ph+height, pw:pw+width] = x.data
        return Tensor(padded_data, requires_grad=x.requires_grad, device=x.device)

    def _im2col(self, x: Tensor, kh: int, kw: int, sh: int, sw: int, out_height: int, out_width: int):
        batch_size, channels, height, width = x.data.shape
        if x.device == 'gpu':
            unfolded_data = mx.zeros((batch_size, out_height * out_width, channels * kh * kw))
        else:
            unfolded_data = np.zeros((batch_size, out_height * out_width, channels * kh * kw))

        for b in range(batch_size):
            col_index = 0;
            for i in range(out_height):
                for j in range(out_width):
                    h_start = i * sh
                    w_start = j * sw
                    window = x.data[b, :, h_start:h_start+kh, w_start:w_start+kw]
                    if x.device == 'gpu':
                        window_flat = window.reshape(-1)
                    else:
                        window_flat = window.ravel()
                    unfolded_data[b, col_index] = window_flat
                    col_index += 1

        return unfolded_data


from energizer.neural_network import Module
from energizer.tensor import Tensor
import numpy as np

class Dropout(Module):
    def __init__(self, p: float = 0.5, inplace: bool = False):
        super().__init__()
        if p < 0 or p > 1:
            raise ValueError("p must be between 0 and 1")
        self.p = p
        self.inplace = inplace
        self.mask = None
        self.training = True

    def forward(self, x: Tensor) -> Tensor:
        if not self.training or self.p == 0:
            return x
        
        scale = 1.0 / (1.0 - self.p)
        mask = (np.random.rand(*x.data.shape) > self.p).astype(np.float32) * scale

        self.mask = mask

        if self.inplace:
            x.data *= mask
            return x
        return Tensor(x.data * mask, requires_grad=x.requires_grad, device=x.device)

    def eval(self):
        self.training = False

    def train(self):
        self.training = True




from energizer.neural_network import Module
from energizer.tensor import Tensor
import numpy as np
import mlx.core as mx

class GELU(Module):
    def __init__(self, device: str = 'cpu'):
        super().__init__(device=device)

    def forward(self, x: Tensor) -> Tensor:
        if x.device == 'gpu':
            return Tensor(0.5 * x.data * (1.0 + mx.tanh(mx.sqrt(2 / mx.pi) * (x.data + 0.044715 * mx.power(x.data, 3)))), requires_grad=x.requires_grad, device=x.device)
        else:
            return Tensor(0.5 * x.data * (1.0 + np.tanh(np.sqrt(2 / np.pi) * (x.data + 0.044715 * np.power(x.data, 3)))), requires_grad=x.requires_grad, device=x.device)



from typing import Callable
from energizer.neural_network import Optimizer
import numpy as np

class Adam(Optimizer):
    def __init__(self, params, lr=0.001, betas=(0.9, 0.999), eps=1e-8, 
                 weight_decay=0, amsgrad=False, maximize=False, **kwargs):

        defaults = dict(lr=lr, betas=betas, eps=eps, 
                       weight_decay=weight_decay, amsgrad=amsgrad, maximize=maximize)
        
        if isinstance(params, (list, tuple)) and len(params) > 0:
            if isinstance(params[0], dict):
                params = params
            else:
                params = [{'params': list(params)}]
        super().__init__(params, defaults)
        for group in self.param_groups:
            for p in group['params']:
                p_data = p.data
                if hasattr(p_data, '__class__') and 'mlx' in str(type(p_data)):
                    import mlx.core as mx
                    exp_avg = mx.zeros_like(p_data)
                    exp_avg_sq = mx.zeros_like(p_data)
                    max_exp_avg_sq = mx.zeros_like(p_data) if group['amsgrad'] else None
                else:
                    exp_avg = np.zeros_like(p_data)
                    exp_avg_sq = np.zeros_like(p_data)
                    max_exp_avg_sq = np.zeros_like(p_data) if group['amsgrad'] else None
                
                self.state[p] = {
                    'step': 0,
                    'exp_avg': exp_avg,
                    'exp_avg_sq': exp_avg_sq
                }
                if group['amsgrad']:
                    self.state[p]['max_exp_avg_sq'] = max_exp_avg_sq

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault('maximize', False)
            group.setdefault('capturable', False)
            group.setdefault('differentiable', False)
            group.setdefault('fused', None)
            group.setdefault('decoupled_decay', False)
            group.setdefault('foreach', False)
            group.setdefault('amsgrad', False)
            group.setdefault('weight_decay', 0)
            group.setdefault('beta1', 0.9)
            group.setdefault('beta2', 0.999)
            group.setdefault('eps', 1e-08)

    def step(self, closure: Callable = None):
        loss = None
        if closure is not None:
            loss = closure()


        for group in self.param_groups:
            lr = group['lr']
            beta1, beta2 = group['betas']
            eps = group['eps']
            weight_decay = group['weight_decay']
            amsgrad = group['amsgrad']
            maximize = group['maximize']

            for p in group['params']:
                if p.grad is None:
                    continue

                grad = p.grad.data if hasattr(p.grad, 'data') else p.grad
                
                if not isinstance(grad, np.ndarray):
                    grad = np.array(grad)
                
                p_data = p.data
                is_mlx = hasattr(p_data, '__class__') and 'mlx' in str(type(p_data))
                if is_mlx and not (hasattr(grad, '__class__') and 'mlx' in str(type(grad))):
                    import mlx.core as mx
                    grad = mx.array(grad)
                elif not is_mlx and (hasattr(grad, '__class__') and 'mlx' in str(type(grad))):
                    grad = np.array(grad)

                if maximize:
                    grad = -grad

                if weight_decay != 0:
                    grad = grad + weight_decay * p_data

                state = self.state[p]
                state['step'] += 1

                state['exp_avg'] = beta1 * state['exp_avg'] + (1 - beta1) * grad
                state['exp_avg_sq'] = beta2 * state['exp_avg_sq'] + (1 - beta2) * grad * grad

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                exp_avg_hat = state['exp_avg'] / bias_correction1
                exp_avg_sq_hat = state['exp_avg_sq'] / bias_correction2

                if amsgrad:
                    if is_mlx:
                        import mlx.core as mx
                        state['max_exp_avg_sq'] = mx.maximum(state['max_exp_avg_sq'], exp_avg_sq_hat)
                        denom = mx.sqrt(state['max_exp_avg_sq']) + eps
                    else:
                        state['max_exp_avg_sq'] = np.maximum(state['max_exp_avg_sq'], exp_avg_sq_hat)
                        denom = np.sqrt(state['max_exp_avg_sq']) + eps
                else:
                    if is_mlx:
                        import mlx.core as mx
                        denom = mx.sqrt(exp_avg_sq_hat) + eps
                    else:
                        denom = np.sqrt(exp_avg_sq_hat) + eps

                p.data = p_data - lr * exp_avg_hat / denom

        return loss



from typing import Callable
import numpy as np
from energizer.neural_network import Optimizer

class SGD(Optimizer):
    def __init__(self, params, lr: float = 0.01, momentum: float = 0.0, dampening: float = 0,  weight_decay: float = 0.0, nesterov: bool = False):
        default = {
            'lr': lr,
            'momentum': momentum,
            'weight_decay': weight_decay,
            'dampening': dampening,
            'nesterov': nesterov
        }
        super().__init__(params, default)

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)
            group.setdefault('dampening', 0)

    def step(self, closure: Callable = None):
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group['lr']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']
            weight_decay = group['weight_decay']

            for p in group['params']:
                if p.grad is None:
                    continue

                grad = p.grad.data if hasattr(p.grad, 'data') else p.grad
                
                p_data = p.data
                is_mlx = hasattr(p_data, '__class__') and 'mlx' in str(type(p_data))
                
                if is_mlx:
                    try:
                        import mlx.core as mx
                        if not (hasattr(grad, '__class__') and 'mlx' in str(type(grad))):
                            grad = mx.array(np.array(grad))
                    except ImportError:
                        grad = np.array(grad)
                else:
                    if not isinstance(grad, np.ndarray):
                        grad = np.array(grad)

                if weight_decay != 0:
                    grad = grad + weight_decay * p.data

                if momentum != 0:
                    if p not in self.state:
                        self.state[p] = {}
                    param_state = self.state[p]
                    if 'momentum_buffer' not in param_state:
                        if is_mlx:
                            import mlx.core as mx
                            buf = param_state['momentum_buffer'] = mx.zeros_like(p_data)
                        else:
                            buf = param_state['momentum_buffer'] = np.zeros_like(p_data)
                    else:
                        buf = param_state['momentum_buffer']
                    buf = momentum * buf + (1 - dampening) * grad
                    param_state['momentum_buffer'] = buf
                    if nesterov:
                        grad = grad + momentum * buf
                    else:
                        grad = buf
                p.data = p_data - lr * grad
        return loss

    



from energizer.neural_network import Module, Parameter
import energizer
import numpy as np
import mlx.core as mx
from energizer.tensor import Tensor

class ConvTranspose2d(Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: tuple, 
                 stride: tuple = 1, padding: tuple = 0, output_padding: tuple = 0, 
                 groups: int = 1, bias: bool = True, dilation: tuple = 1,
                 device: str = 'cpu'):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
        self.stride = stride if isinstance(stride, tuple) else (stride, stride)
        self.padding = padding if isinstance(padding, tuple) else (padding, padding)
        self.output_padding = output_padding if isinstance(output_padding, tuple) else (output_padding, output_padding)
        self.groups = groups
        self.bias = bias
        self.dilation = dilation if isinstance(dilation, tuple) else (dilation, dilation)
        self.device = device

        weight_shape = (in_channels, out_channels, self.kernel_size[0], self.kernel_size[1])
        
        if device == 'gpu':
            weight_data = mx.random.normal(weight_shape) * 0.01
            bias_data = mx.zeros(out_channels) if bias else None
        else:
            weight_data = np.random.randn(*weight_shape) * 0.01
            bias_data = np.zeros(out_channels) if bias else None
        
        self.weight = Parameter(
            weight_data,
            requires_grad=True,
            device=device
        )
        
        if bias:
            self.bias = Parameter(bias_data, requires_grad=True, device=device)
        else:
            self.bias = None

    def forward(self, x: Tensor) -> Tensor:
        if x.device != self.device:
            x = x.to(self.device)
        
        batch_size, in_channels, height, width = x.shape

        kh, kw = self.kernel_size
        sh, sw = self.stride
        ph, pw = self.padding
        oph, opw = self.output_padding
        dh, dw = self.dilation

        if (dh, dw) != (1, 1):
            raise NotImplementedError("ConvTranspose2d dilation != 1 is not supported yet")

        if self.device == 'gpu':
            if not isinstance(x.data, mx.array):
                x_data = mx.array(x.data)
            else:
                x_data = x.data
            
            if not isinstance(self.weight.data, mx.array):
                weight_data = mx.array(self.weight.data)
            else:
                weight_data = self.weight.data
            
            x_channels_last = mx.transpose(x_data, (0, 2, 3, 1))
            
            weight_mlx = mx.transpose(weight_data, (1, 2, 3, 0))

            kwargs = {
                "stride": (sh, sw),
                "padding": (ph, pw),
                "dilation": (dh, dw),
                "groups": self.groups   
            }
            if (oph, opw) != (0, 0):
                kwargs["output_padding"] = (oph, opw)
            
            output = mx.conv_transpose2d(x_channels_last, weight_mlx, **kwargs)
            output = mx.transpose(output, (0, 3, 1, 2))
        else:
            x_data = x.data if isinstance(x.data, np.ndarray) else np.array(x.data)
            weight_data = self.weight.data if isinstance(self.weight.data, np.ndarray) else np.array(self.weight.data)
            
            out_height = (height - 1) * sh - 2 * ph + (kh - 1) + oph + 1
            out_width = (width - 1) * sw - 2 * pw + (kw - 1) + opw + 1

            up_h = (height - 1) * sh + 1
            up_w = (width - 1) * sw + 1
            upsampled = np.zeros((batch_size, in_channels, up_h, up_w), dtype=x_data.dtype)
            upsampled[:, :, ::sh, ::sw] = x_data

            pad_h = kh - 1 - ph
            pad_w = kw - 1 - pw
            padded = np.pad(
                upsampled,
                ((0, 0), (0, 0), (pad_h, pad_h), (pad_w, pad_w)),
                mode='constant'
            )

            flipped_weight = np.flip(weight_data, axis=(2, 3))
            conv_h = padded.shape[2] - kh + 1
            conv_w = padded.shape[3] - kw + 1
            
            output = np.zeros((batch_size, self.out_channels, conv_h, conv_w), dtype=x_data.dtype)
            
            for b in range(batch_size):
                for oc in range(self.out_channels):
                    for ic in range(in_channels):
                        kernel = flipped_weight[ic, oc]
                        for i in range(conv_h):
                            for j in range(conv_w):
                                patch = padded[b, ic, i:i+kh, j:j+kw]
                                output[b, oc, i, j] += np.sum(patch * kernel)
            
            if oph or opw:
                output = np.pad(output, ((0, 0), (0, 0), (0, oph), (0, opw)), mode='constant')

            output = output[:, :, :out_height, :out_width]

        if self.bias is not None:
            if self.device == 'gpu':
                if isinstance(output, mx.array):
                    bias_data = self.bias.mlx() if hasattr(self.bias, 'mlx') else mx.array(self.bias.data)
                    output += bias_data.reshape(1, -1, 1, 1)
                else:
                    output += np.array(self.bias.data).reshape(1, -1, 1, 1)
            else:
                bias_data = self.bias.data if isinstance(self.bias.data, np.ndarray) else np.array(self.bias.data)
                output += bias_data.reshape(1, -1, 1, 1)

        return Tensor(output, requires_grad=x.requires_grad, device=self.device)
    
    def to(self, device: str):
        self.device = device
        self.weight = self.weight.to(device)
        if self.bias is not None:
            self.bias = self.bias.to(device)
        return self


import numpy as np
import mlx.core as mx
from typing import Union, Optional

import energizer.derivatives as dv
from .function import Function

class Tensor:
    """
    A tensor is a multi-dimensional array.
    """
    def __init__(self, data, requires_grad: bool = False, grad_fn: Optional["Function"] = None, device: str = 'cpu'):
        self.device = device
        self.requires_grad = requires_grad
        self.grad_fn = grad_fn
        self.grad = None

        if self.device == 'cpu':
            self._init_cpu(data)
        elif self.device == 'gpu':
            self._init_mlx(data)

    def __str__(self):
        # Simple debugger-style representation showing device and backend
        backend = "mlx" if isinstance(self.data, mx.array) else "numpy"
        data_str = np.array2string(
            np.array(self.data),
            separator=', ',
            precision=4,
            suppress_small=False,
            floatmode='maxprec_equal'
        )
        return (
            f"tensor({data_str}, "
            f"device={self.device}, "
            f"backend={backend}, "
            f"requires_grad={self.requires_grad})"
        )

    def _init_cpu(self, data):
        if isinstance(data, np.ndarray):
            self.data = data
        elif isinstance(data, mx.array):
            self.data = np.array(data)
        elif isinstance(data, Tensor):
            if data.device == 'gpu':
                self.data = np.array(data.data)
            else:
                self.data = data.data.copy()
        else:
            self.data = np.array(data)

        self.shape = self.data.shape
    
    def _init_mlx(self, data):
        if isinstance(data, mx.array):
            self.data = data
        elif isinstance(data, np.ndarray):
            self.data = mx.array(data)
        elif isinstance(data, Tensor):
            if data.device == 'gpu':
                self.data = data.data
            else:
                self.data = mx.array(data.data)
        else:
            self.data = mx.array(data)
        
        self.shape = self.data.shape
        self.device = 'gpu'

    def cpu(self) -> 'Tensor':
        if self.device == 'cpu':
            return self
        return Tensor(np.array(self.data), requires_grad=self.requires_grad, device='cpu')

    def gpu(self) -> 'Tensor':
        if self.device == 'gpu':
            return self
        return Tensor(mx.array(self.data), requires_grad=self.requires_grad, device='gpu')

    def to(self, device: str) -> 'Tensor':
        if self.device == device:
            return self
        if device == 'cpu':
            return self.cpu()
        elif device == 'gpu':
            return self.gpu()
        else:
            raise ValueError(f"Invalid device: {device}")

    def numpy(self) -> np.ndarray:
        if isinstance(self.data, mx.array):
            return np.array(self.data)
        return self.data.copy()

    def mlx(self) -> mx.array:
        if isinstance(self.data, mx.array):
            return self.data
        return mx.array(self.data)

    def _get_other_data(self, other: Union[int, float, 'Tensor']):
        if isinstance(other, (int, float)):
            return other
        elif isinstance(other, Tensor):
            if self.device == 'cpu':
                return other.cpu().data
            elif self.device == 'gpu':
                return other.gpu().data
        # Handle MLX arrays and numpy arrays directly
        elif isinstance(other, mx.array):
            if self.device == 'cpu':
                return np.array(other)
            else:
                return other
        elif isinstance(other, np.ndarray):
            if self.device == 'gpu':
                return mx.array(other)
            else:
                return other
        # Fallback: try to access .data if it exists
        elif hasattr(other, 'data'):
            return other.data
        else:
            return other

    def requires_grad_(self, requires_grad: bool = True):
        self.requires_grad = requires_grad
        return self

    def __mul__(self, other: Union[int, float, 'Tensor']) -> 'Tensor':
        other_data = self._get_other_data(other)
        return Tensor(self.data * other_data, requires_grad=self.requires_grad, grad_fn=Function(dv.mul_backward, [self, other]), device=self.device)
    
    def __rmul__(self, other: Union[int, float]) -> 'Tensor':
        return Tensor(other * self.data, requires_grad=self.requires_grad, grad_fn=Function(dv.mul_backward, [self, other]), device=self.device)

    def __add__(self, other: Union[int, float, 'Tensor']) -> 'Tensor':
        other_data = self._get_other_data(other)
        return Tensor(self.data + other_data, requires_grad=self.requires_grad, grad_fn=Function(dv.add_backward, [self, other]), device=self.device)

    def __sub__(self, other: Union[int, float, 'Tensor']) -> 'Tensor':
        other_data = self._get_other_data(other)
        return Tensor(self.data - other_data, requires_grad=self.requires_grad, grad_fn=Function(dv.sub_backward, [self, other]), device=self.device)

    def __neg__(self) -> 'Tensor':
        return Tensor(isinstance(self.data, (int, float)) -self.data if isinstance(self.data, (int, float)) else -self.data, requires_grad=self.requires_grad, grad_fn=Function(dv.neg_backward, [self]), device=self.device)

    def __truediv__(self, other: Union[int, float, 'Tensor']) -> 'Tensor':
        other_data = self._get_other_data(other)
        return Tensor(self.data / other_data, requires_grad=self.requires_grad, grad_fn=Function(dv.truediv_backward, [self, other]), device=self.device)

    def __rtruediv__(self, other: Union[int, float]) -> 'Tensor':
        other = self._get_other_data(other)
        return Tensor(other / self.data, requires_grad=self.requires_grad, grad_fn=Function(dv.truediv_backward, [other, self]), device=self.device)

    def __matmul__(self, other: 'Tensor') -> 'Tensor':     
        other_data = self._get_other_data(other)
        return Tensor(
            self.data @ other_data, 
            requires_grad=self.requires_grad or other.requires_grad,
            grad_fn=Function(dv.matmul_backward, [self, other]),
            device=self.device
        )

    def __pow__(self, other: Union[int, float]) -> 'Tensor':
        return Tensor(self.data ** other, requires_grad=self.requires_grad, grad_fn=Function(dv.pow_backward, [self, other]), device=self.device)

    def __rmatmul__(self, other: 'Tensor') -> 'Tensor':
        return self.__matmul__(other)

    def __getitem__(self, key):
        return Tensor(self.data[key], requires_grad=self.requires_grad, grad_fn=Function(dv.getitem_backward, [self, key]), device=self.device)

    def __setitem__(self, key, value: Union[int, float, 'Tensor']):
        if isinstance(value, Tensor):
            value_data = value.data
        else:
            value_data = value

        if isinstance(self.data, mx.array):
            temp_np = np.array(self.data)
            temp_np[key] = value_data
            new_data = mx.array(temp_np)
        else:
            new_data = self.data.copy()
            new_data[key] = value_data

        if self.requires_grad:
            self.grad_fn = Function(dv.setitem_backward, [self, key, value])

        self.data = new_data

    @property
    def T(self) -> 'Tensor':
        return Tensor(self.data.T, requires_grad=self.requires_grad, grad_fn=Function(dv.transpose_backward, [self]), device=self.device)

    @property
    def size(self):
        return self.data.shape

    def item(self) -> Union[int, float]:
        if self.data.size != 1:
            raise ValueError("Tensor must be a single element")

        scalar_value = self.data.item()
        if self.requires_grad:
            self.grad_fn = Tensor(scalar_value, requires_grad=self.requires_grad, grad_fn=Function(dv.item_backward, [self]), device=self.device)
        return scalar_value

    def copy(self) -> 'Tensor':
        return Tensor(self.data.copy(), requires_grad=self.requires_grad, grad_fn=self.grad_fn)

    def sum(self, dim: Optional[int] = None) -> 'Tensor':
        if isinstance(self.data, mx.array):
            result = mx.sum(self.data, axis=dim)
        else:
            result = np.sum(self.data, axis=dim)
        return Tensor(result, requires_grad=self.requires_grad, grad_fn=Function(dv.sum_backward, [self, dim]), device=self.device)

    def mean(self, dim: Optional[int] = None) -> 'Tensor':
        if isinstance(self.data, mx.array):
            result = mx.mean(self.data, axis=dim)   
        else:
            result = np.mean(self.data, axis=dim)
        return Tensor(result, requires_grad=self.requires_grad, grad_fn=Function(dv.mean_backward, [self, dim]), device=self.device)

    def reshape(self, shape: tuple) -> 'Tensor':
        if isinstance(self.data, mx.array):
            result = self.data.reshape(shape)
        else:
            result = self.data.reshape(shape)
        return Tensor(result, requires_grad=self.requires_grad, grad_fn=Function(dv.reshape_backward, [self, shape]), device=self.device)

    def view(self, shape: tuple) -> 'Tensor':
        return self.reshape(shape)

    def transpose(self, dim0: int, dim1: int) -> 'Tensor':
        if self.device == 'gpu':
            result = mx.swapaxes(self.data, dim0, dim1)
        else:
            result = np.swapaxes(self.data, dim0, dim1)
        return Tensor(result, requires_grad=self.requires_grad, grad_fn=Function(dv.transpose_backward, [self, dim0, dim1]), device=self.device)

    @staticmethod
    def randn(*args, device: str = 'cpu', **kwargs):
        if device == 'gpu':
            data = mx.random.normal(shape=args, **kwargs)
        else:
            data = np.random.randn(*args, **kwargs)

        return Tensor(data, requires_grad=False, device=device)

    @staticmethod
    def zeros(shape, device: str = 'cpu'):
        if device == 'gpu':
            data = mx.zeros(shape)
        else:
            data = np.zeros(shape)
        return Tensor(data, requires_grad=False, device=device)

    @staticmethod
    def ones(shape, device: str = 'cpu'):
        if device == 'gpu':
            data = mx.ones(shape)
        else:
            data = np.ones(shape)
        return Tensor(data, requires_grad=False, device=device)

    def backward(self, grad_outputs = None):
        if grad_outputs is None:
            if isinstance(self.data, mx.array):
                grad_outputs = Tensor(mx.ones_like(self.data), device=self.device)
            else:
                grad_outputs = Tensor(np.ones_like(self.data), device=self.device)
        else:
            # Convert numpy/mx array to Tensor if needed
            if not isinstance(grad_outputs, Tensor):
                grad_outputs = Tensor(grad_outputs, device=self.device)
            # Ensure grad_outputs is on the same device as self
            elif self.device != grad_outputs.device:
                grad_outputs = grad_outputs.to(self.device)
        
        if self.requires_grad and self.grad_fn is not None:
            self.grad_fn.backward([grad_outputs])
    

def tensor(data, requires_grad=False, device='cpu'):
    return Tensor(data, requires_grad=requires_grad, device=device)